import textwrap
from typing import List, TYPE_CHECKING
from collections.abc import Sequence

import pytest

from debputy.lsp.languages.lsp_debian_control import _reformat_debian_control
from debputy.lsp.languages.lsp_debian_watch import _reformat_debian_watch
from debputy.lsp.maint_prefs import MaintainerPreferenceTable
from debputy.lsp.text_edit import apply_text_edits
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from lint_tests.lint_tutil import (
    ReformatWrapper,
    apply_formatting_edits,
)

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


@pytest.fixture
def reformater(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lint_dctrl_parser: DctrlParser,
    maintainer_preference_table: MaintainerPreferenceTable,
) -> ReformatWrapper:

    return ReformatWrapper(
        "/nowhere/debian/watch",
        _reformat_debian_watch,
        debputy_plugin_feature_set,
        lint_dctrl_parser,
        maintainer_preference_table,
    )


def test_dwatch_reformat(reformater: ReformatWrapper) -> None:
    lines = textwrap.dedent("""\
        Version: 5

        Source: http://example.com/release/@PACKAGE@.html
        # This comment should be preserved
        MatchingPattern: files/@PACKAGE@@ANY_VERSION@@ARCHIVE_EXT@
        # This comment should be preserved
        Pgp-SigUrlMangle: s%@ARCHIVE_EXT@$%.asc%
        Decompression: yes
    """).splitlines(keepends=True)

    edits = reformater.reformat(lines)
    # By default, we do nothing
    assert not edits

    edits_black = reformater.reformat_with_named_style("black", lines)
    assert edits_black

    actual_reformatted_black = apply_formatting_edits(lines, edits_black)

    expected_reformatted_black = textwrap.dedent("""\
        Version: 5

        Source: http://example.com/release/@PACKAGE@.html
        # This comment should be preserved
        Matching-Pattern: files/@PACKAGE@@ANY_VERSION@@ARCHIVE_EXT@
        # This comment should be preserved
        Pgp-Sig-Url-Mangle: s%@ARCHIVE_EXT@$%.asc%
        Decompression: yes
    """)

    assert actual_reformatted_black == expected_reformatted_black



from pathlib import Path

import pytest

from debputy.util import compute_output_filename


def write_unpacked_deb(
    root: Path,
    package: str,
    version: str,
    arch: str,
    arch_variant: str | None,
) -> None:
    lines = [
        f"Package: {package}",
        f"Version: {version}",
        f"Architecture: {arch}",
    ]
    if arch_variant:
        lines.append(f"Architecture-Variant: {arch_variant}")
    (root / "control").write_text("".join(f"{x}\n" for x in lines))


@pytest.mark.parametrize(
    "package,version,arch,is_udeb,arch_variant,expected",
    [
        ("fake", "1.0", "amd64", False, None, "fake_1.0_amd64.deb"),
        ("fake", "1.0", "amd64", True, None, "fake_1.0_amd64.udeb"),
        ("fake", "2:1.0", "amd64", False, None, "fake_1.0_amd64.deb"),
        ("fake", "2:1.0", "amd64", True, None, "fake_1.0_amd64.udeb"),
        ("fake", "3.0", "all", False, None, "fake_3.0_all.deb"),
        ("fake", "3.0", "all", True, None, "fake_3.0_all.udeb"),
        # Architecture-Variant
        ("fake", "1.0", "amd64", False, "i386", "fake_1.0_i386.deb"),
        ("fake", "1.0", "amd64", True, "i386", "fake_1.0_i386.udeb"),
    ],
)
def test_generate_deb_filename(
    tmp_path: Path,
    package: str,
    version: str,
    arch: str,
    is_udeb: bool,
    arch_variant: str | None,
    expected: str,
) -> None:
    write_unpacked_deb(tmp_path, package, version, arch, arch_variant)
    assert compute_output_filename(str(tmp_path), is_udeb) == expected



from typing import (
    Optional,
    Any,
    Tuple,
    Union,
)
from collections.abc import Sequence, Container

from debputy.linting.lint_util import LintState, DebputyMetadata
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_standard_handler,
    lsp_hover,
    lsp_completer,
    SecondaryLanguage,
    LanguageDispatchRule,
)
from debputy.lsp.lsp_generic_yaml import (
    yaml_key_range,
    generic_yaml_hover,
    LSPYAMLHelper,
    generic_yaml_lint,
    generic_yaml_completer,
    _is_empty_range,
)
from debputy.lsprotocol.types import (
    TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL,
    HoverParams,
    Hover,
    TEXT_DOCUMENT_CODE_ACTION,
    CompletionParams,
    CompletionList,
    CompletionItem,
)
from debputy.plugin.api.impl_types import (
    PluginProvidedParser,
)
from debputy.plugin.api.parser_tables import OPARSER_MANIFEST_ROOT
from debputy.plugin.api.spec import DebputyIntegrationMode
from debputy.plugins.debputy.private_api import Capability, load_libcap

try:
    from pygls.server import LanguageServer
    from debputy.lsp.debputy_ls import DebputyLanguageServer
except ImportError:
    pass


_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/debputy.manifest",
    "debputy.manifest",
    "debian/debputy.manifest",
    [
        SecondaryLanguage("debputy.manifest"),
        # LSP's official language ID for YAML files
        SecondaryLanguage("yaml", secondary_lookup="path-name"),
    ],
)


lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)


class DebputyManifestLSPYAMLHelper(LSPYAMLHelper[DebputyMetadata]):

    def _validate_subparser_is_valid_here(
        self,
        subparser: PluginProvidedParser,
        orig_key: str,
        line: int,
        col: int,
    ) -> None:
        _integration_mode_allows_key(
            self.lint_state,
            self.custom_data.debputy_integration_mode,
            subparser.parser.expected_debputy_integration_mode,
            orig_key,
            line,
            col,
        )

    def _type_based_value_check(
        self,
        target_attr_type: type,
        value: Any,
        value_pos: tuple[int, int],
        *,
        key: str | int | None = None,
    ) -> bool:
        if issubclass(target_attr_type, Capability):
            has_libcap, _, is_valid_cap = load_libcap()
            if has_libcap and not is_valid_cap(value):
                line_no, cursor_pos = value_pos
                cap_range = self._remaining_line(line_no, cursor_pos)
                if _is_empty_range(cap_range):
                    # FIXME: We cannot report an empty range, but there is still a problem here.
                    return True
                self.lint_state.emit_diagnostic(
                    cap_range,
                    "The value could not be parsed as a capability via cap_from_text on this system",
                    "warning",
                    "debputy",
                )
            return True

        return super()._type_based_value_check(
            target_attr_type,
            value,
            value_pos,
            key=key,
        )


def _initialize_yaml_helper(lint_state: LintState) -> DebputyManifestLSPYAMLHelper:
    return DebputyManifestLSPYAMLHelper(
        lint_state,
        lint_state.plugin_feature_set.manifest_parser_generator,
        lint_state.debputy_metadata,
    )


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_debputy_manifest(lint_state: LintState) -> None:
    pg = lint_state.plugin_feature_set.manifest_parser_generator
    root_parser = pg.dispatchable_object_parsers[OPARSER_MANIFEST_ROOT]
    await generic_yaml_lint(
        lint_state,
        root_parser,
        _initialize_yaml_helper,
    )


def _integration_mode_allows_key(
    lint_state: LintState,
    debputy_integration_mode: DebputyIntegrationMode | None,
    expected_debputy_integration_modes: Container[DebputyIntegrationMode] | None,
    key: str,
    line: int,
    col: int,
) -> None:
    if debputy_integration_mode is None or expected_debputy_integration_modes is None:
        return
    if debputy_integration_mode in expected_debputy_integration_modes:
        return
    key_range = yaml_key_range(key, line, col)
    lint_state.emit_diagnostic(
        key_range,
        f'Feature "{key}" not supported in integration mode {debputy_integration_mode}',
        "error",
        "debputy",
    )


@lsp_completer(_DISPATCH_RULE)
def debputy_manifest_completer(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
) -> CompletionList | Sequence[CompletionItem] | None:
    pg = ls.plugin_feature_set.manifest_parser_generator
    root_parser = pg.dispatchable_object_parsers[OPARSER_MANIFEST_ROOT]
    return generic_yaml_completer(
        ls,
        params,
        root_parser,
    )


@lsp_hover(_DISPATCH_RULE)
def debputy_manifest_hover(
    ls: "DebputyLanguageServer",
    params: HoverParams,
) -> Hover | None:
    return generic_yaml_hover(
        ls,
        params,
        lambda pg: pg.dispatchable_object_parsers[OPARSER_MANIFEST_ROOT],
        show_integration_mode=True,
    )



import textwrap
from typing import TYPE_CHECKING

import pytest

from debputy.lsp.languages.lsp_debian_changelog import _lint_debian_changelog
from debputy.lsprotocol.types import DiagnosticSeverity
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from lint_tests.lint_tutil import (
    LintWrapper,
    diag_range_to_text,
    standard_quick_fixes_for_diagnostic,
    apply_code_action_edits,
    code_actions_for,
)

if TYPE_CHECKING:
    from lsprotocol import types
else:
    from debputy.lsprotocol import types


@pytest.fixture
def line_linter(
    debputy_plugin_feature_set: PluginProvidedFeatureSet,
    lint_dctrl_parser: DctrlParser,
) -> LintWrapper:
    return LintWrapper(
        "/nowhere/debian/changelog",
        _lint_debian_changelog,
        debputy_plugin_feature_set,
        lint_dctrl_parser,
    )


def test_dch_lint(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent("""\
    foo (0.2) unstable; urgency=medium

      * Renamed to foo
    
     -- Niels Thykier <niels@thykier.net>  Mon, 08 Apr 2024 16:00:00 +0000

    bar (0.2) unstable; urgency=medium

      * Initial release
    
     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    print(diagnostics)
    # Without a control file, this is fine
    assert not diagnostics

    line_linter.dctrl_lines = textwrap.dedent("""\
    Source: foo
    
    Package: something-else
    """).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    print(diagnostics)
    # Also fine, because d/control and d/changelog agrees
    assert not diagnostics

    line_linter.dctrl_lines = textwrap.dedent("""\
    Source: bar

    Package: something-else
    """).splitlines(keepends=True)

    diagnostics = line_linter(lines)
    print(diagnostics)
    # This should be problematic though
    assert diagnostics and len(diagnostics) == 1
    diag = diagnostics[0]

    msg = (
        "The first entry must use the same source name as debian/control."
        ' Changelog uses: "foo" while d/control uses: "bar"'
    )
    assert diag.severity == DiagnosticSeverity.Error
    assert diag.message == msg
    assert f"{diag.range}" == "0:0-0:3"


def test_dch_lint_historical(line_linter: LintWrapper) -> None:
    nonsense = "very very very very very very very very very very very very very very "
    lines = textwrap.dedent(f"""\
    foo (0.4) unstable; urgency=medium

      * A {nonsense} long line about absolute nothing that should trigger a warning about length.

     -- Niels Thykier <niels@thykier.net>  Mon, 08 Apr 2024 16:00:00 +0000

    foo (0.3) unstable; urgency=medium

      * Another entry that is not too long.

     -- Niels Thykier <niels@thykier.net>  Thu, 04 Apr 2024 00:00:00 +0000

    foo (0.2) unstable; urgency=medium

      * A {nonsense}  long line about absolute nothing that should not trigger a warning about length.

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    # This should be problematic though
    assert diagnostics and len(diagnostics) == 1
    diag = diagnostics[0]

    msg = "Line exceeds 81 characters"
    assert diag.severity == DiagnosticSeverity.Hint
    assert diag.message == msg
    assert f"{diag.range}" == "2:80-2:153"


def test_dch_lint_invalid_version(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(f"""\
    foo (a!0.3!a) unstable; urgency=medium

      * Another entry that is not too long.

     -- Niels Thykier <niels@thykier.net>  Thu, 04 Apr 2024 00:00:00 +0000

    foo (a+!) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    assert diagnostics and len(diagnostics) == 3
    first_issue, second_issue, third_issue = diagnostics

    msg = "This part cannot be parsed as a valid Debian version"
    assert first_issue.severity == DiagnosticSeverity.Error
    assert first_issue.message == msg
    assert f"{first_issue.range}" == "0:5-0:7"

    msg = "This part cannot be parsed as a valid Debian version"
    assert second_issue.severity == DiagnosticSeverity.Error
    assert second_issue.message == msg
    assert f"{second_issue.range}" == "0:10-0:12"

    msg = 'Cannot parse "a+!" as a Debian version.'
    assert third_issue.severity == DiagnosticSeverity.Error
    assert third_issue.message == msg
    assert f"{third_issue.range}" == "6:5-6:8"


def test_dch_lint_version_typo_dfsg(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(f"""\
    foo (0.3+dsfg) unstable; urgency=medium

      * Another entry that is not too long.

     -- Niels Thykier <niels@thykier.net>  Thu, 04 Apr 2024 00:00:00 +0000

    foo (0.2) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    assert diagnostics and len(diagnostics) == 1
    issue = diagnostics[0]

    msg = 'Typo of "dfsg" (Debian Free Software Guidelines)'
    assert issue.severity == DiagnosticSeverity.Hint
    assert issue.message == msg
    assert f"{issue.range}" == "0:9-0:13"
    assert diag_range_to_text(lines, issue.range) == "dsfg"


@pytest.mark.parametrize(
    "spacing,expected_range,highlighted_text",
    [
        (
            "",
            "4:36-4:38",
            ">T",
        ),
        (
            " ",
            "4:37-4:38",
            " ",
        ),
        (
            "\t\t",
            "4:37-4:39",
            "\t\t",
        ),
        (
            "   ",
            "4:37-4:40",
            "   ",
        ),
    ],
)
def test_dch_lint_signoff_line_email_date_spacing(
    line_linter: LintWrapper,
    spacing: str,
    expected_range: str,
    highlighted_text: str,
) -> None:
    lines = textwrap.dedent(f"""\
    foo (0.3) unstable; urgency=medium

      * Another entry that is not too long.

     -- Niels Thykier <niels@thykier.net>{spacing}Thu, 04 Apr 2024 00:00:00 +0000

    foo (0.2) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    assert diagnostics and len(diagnostics) == 1
    issue = diagnostics[0]

    msg = "Must be exactly two spaces between email and sign off date"
    assert issue.severity == DiagnosticSeverity.Error
    assert issue.message == msg
    assert f"{issue.range}" == expected_range
    assert diag_range_to_text(lines, issue.range) == highlighted_text
    quick_fixes = standard_quick_fixes_for_diagnostic(
        line_linter.lint_state,
        code_actions_for(line_linter.lint_state.doc_uri, issue.range),
        issue,
    )
    assert len(quick_fixes) == 1
    quick_fix = quick_fixes[0]
    assert isinstance(quick_fix, types.CodeAction)
    fixed_text = apply_code_action_edits(
        line_linter.lint_state.doc_uri,
        lines,
        quick_fix,
    )
    fixed_lines = fixed_text.splitlines(keepends=True)
    expected_line_post_fix = (
        " -- Niels Thykier <niels@thykier.net>  Thu, 04 Apr 2024 00:00:00 +0000\n"
    )
    assert fixed_lines[4] == expected_line_post_fix


@pytest.mark.parametrize(
    "spacing",
    [
        "",
        " ",
        "   ",
    ],
)
def test_dch_lint_signoff_missing_sign_off_date(
    line_linter: LintWrapper,
    spacing: str,
) -> None:
    lines = textwrap.dedent(f"""\
    foo (0.3) unstable; urgency=medium

      * Another entry that is not too long.

     -- Niels Thykier <niels@thykier.net>{spacing}

    foo (0.2) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    assert diagnostics and len(diagnostics) == 1
    issue = diagnostics[0]

    msg = "Missing sign off date"
    assert issue.severity == DiagnosticSeverity.Error
    assert issue.message == msg
    # The line was always stripped when the test was written, so the trailing space
    # was never included in the range for that reason.
    assert f"{issue.range}" == f"4:0-4:37"
    expected_text = f" -- Niels Thykier <niels@thykier.net>"
    assert diag_range_to_text(lines, issue.range) == expected_text


def test_dch_lint_signoff_empty_sign_off(line_linter: LintWrapper) -> None:
    lines = textwrap.dedent(f"""\
    foo (0.3) unstable; urgency=medium

      * Another entry that is not too long.

     --

    foo (0.2) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    assert diagnostics and len(diagnostics) == 1
    issue = diagnostics[0]

    msg = 'Missing "Name <email@example.com>"'
    assert issue.severity == DiagnosticSeverity.Error
    assert issue.message == msg
    assert f"{issue.range}" == f"4:0-4:3"
    expected_text = f" --"
    assert diag_range_to_text(lines, issue.range) == expected_text

    lines = textwrap.dedent(f"""\
    foo (0.3) unstable; urgency=medium

      * Another entry that is not too long.

     --Niels

    foo (0.2) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(diagnostics)
    assert diagnostics and len(diagnostics) == 1
    issue = diagnostics[0]

    msg = 'Start of sign-off line should be " -- ".'
    assert issue.severity == DiagnosticSeverity.Error
    assert issue.message == msg
    assert f"{issue.range}" == f"4:0-4:8"
    expected_text = f" --Niels"
    assert diag_range_to_text(lines, issue.range) == expected_text


@pytest.mark.parametrize(
    "signoff_line,issue_range,severity,issue_msg,highlight_text,ignore_extra_diagnostics",
    [
        (
            " -- Niels Thykier niels@thykier.net  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:4-4:35",
            types.DiagnosticSeverity.Error,
            'Missing opening "<" to start the email address after the name',
            "Niels Thykier niels@thykier.net",
            False,
        ),
        (
            " -- Niels Thykier niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:18-4:36",
            types.DiagnosticSeverity.Error,
            'Missing opening "<" to start the email address after the name',
            "niels@thykier.net>",
            False,
        ),
        (
            " -- Niels Thykier niels@thykier.net Mon, 01 Apr 2024 00:00:00 +0000",
            "4:4-4:67",
            types.DiagnosticSeverity.Error,
            'Missing opening "<" to start the email address after the name',
            "Niels Thykier niels@thykier.net Mon, 01 Apr 2024 00:00:00 +0000",
            False,
        ),
        (
            " -- Niels Thykier <niels@thykier.net Mon, 01 Apr 2024 00:00:00 +0000",
            "4:18-4:36",
            types.DiagnosticSeverity.Error,
            'Missing closing ">" to finish email address before the sign off date',
            "<niels@thykier.net",
            # This also triggers the missing "  " spaces between email and date.
            # The point is deliberately to test that a single space is sufficient
            # to delimit the email, so we cannot avoid the second error here.
            True,
        ),
        (
            " -- <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:3-4:4",
            types.DiagnosticSeverity.Error,
            "Missing name before email",
            " ",
            False,
        ),
        (
            " --  <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:4-4:5",
            types.DiagnosticSeverity.Error,
            "Missing name before email",
            " ",
            False,
        ),
        (
            " --   Niels<niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:4-4:11",
            types.DiagnosticSeverity.Warning,
            "Non-standard spacing around the name",
            "  Niels",
            False,
        ),
        (
            " -- Niels  <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:4-4:11",
            types.DiagnosticSeverity.Warning,
            "Non-standard spacing around the name",
            "Niels  ",
            False,
        ),
        (
            " --  Niels <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000",
            "4:4-4:11",
            types.DiagnosticSeverity.Warning,
            "Non-standard spacing around the name",
            " Niels ",
            False,
        ),
        (
            " -- Niels Thykier <niels@thykier.net>  Mon",
            "4:39-4:42",
            types.DiagnosticSeverity.Error,
            "Expected a date in RFC822 format (Tue, 12 Mar 2024 12:34:56 +0000)",
            "Mon",
            False,
        ),
        (
            " -- Niels Thykier <niels@thykier.net>  ???, 01 Apr 2024 00:00:00 +0000",
            "4:39-4:42",
            types.DiagnosticSeverity.Error,
            "Expected a three letter date here using US English format (Mon, Tue, ..., Sun)",
            "???",
            False,
        ),
        (
            " -- Niels Thykier <niels@thykier.net>  Mon 2025-05-01",
            "4:42-4:44",
            types.DiagnosticSeverity.Error,
            'Improper formatting of date. Expected ", " here, not " 2"',
            " 2",
            False,
        ),
        (
            " -- Niels Thykier <niels@thykier.net>  Mon, Golf",
            "4:44-4:48",
            types.DiagnosticSeverity.Error,
            'Unable to parse the date as a valid RFC822 date: Invalid date value or format "Golf"',
            "Golf",
            False,
        ),
        (
            " -- Niels Thykier <niels@thykier.net>  Tue, 01 Apr 2024 00:00:00 +0000",
            "4:39-4:42",
            types.DiagnosticSeverity.Warning,
            "The date was a Monday",
            "Tue",
            False,
        ),
    ],
)
def test_dch_lint_signoff_problems(
    line_linter: LintWrapper,
    signoff_line: str,
    issue_range: str,
    severity: "types.DiagnosticSeverity",
    issue_msg: str,
    highlight_text: str,
    ignore_extra_diagnostics: bool,
) -> None:
    lines = textwrap.dedent("""\
    foo (0.3) unstable; urgency=medium

      * Another entry that is not too long.

    {signoff_line}

    foo (0.2) unstable; urgency=medium

      * Initial release

     -- Niels Thykier <niels@thykier.net>  Mon, 01 Apr 2024 00:00:00 +0000
    """).format(signoff_line=signoff_line).splitlines(keepends=True)
    diagnostics = line_linter(lines)
    print(lines)
    print(diagnostics)
    assert diagnostics
    assert len(diagnostics) == 1 or ignore_extra_diagnostics
    issue = diagnostics[0]

    assert issue.severity == severity
    assert issue.message == issue_msg
    assert f"{issue.range}" == issue_range
    assert diag_range_to_text(lines, issue.range) == highlight_text



import dataclasses
import textwrap


@dataclasses.dataclass(slots=True, frozen=True)
class PublicNamedStyle:
    name: str
    long_description: str


ALL_PUBLIC_NAMED_STYLES = {
    s.name: s
    for s in [
        PublicNamedStyle(
            "black",
            long_description=textwrap.dedent("""\
                Uncompromising file formatting of Debian packaging files
        
                By using it, you  agree to cede control over minutiae of hand-formatting. In
                return, the formatter gives you speed, determinism, and freedom from style
                discussions about formatting.
        
                The `black` style is inspired by the `black` Python code formatter. Like with
                `black`, the style will evolve over time.
            """),
        ),
    ]
}



from . import lsp_debian_changelog
from . import lsp_debian_control
from . import lsp_debian_copyright
from . import lsp_debian_debputy_manifest
from . import lsp_debian_patches_series
from . import lsp_debian_rules
from . import lsp_debian_tests_control
from . import lsp_debian_upstream_metadata
from . import lsp_debian_watch
from . import lsp_debputy_config



#!/usr/bin/python3 -B
import pathlib
import sys

# Replaced during install; must be a single line
# fmt: off
DEBPUTY_ROOT_DIR = pathlib.Path(__file__).parent / 'src'
# fmt: on

if __name__ == "__main__":
    # setup PYTHONPATH: add our installation directory.
    sys.path.insert(0, str(DEBPUTY_ROOT_DIR))
    from debputy.commands.deb_materialization import main

    main()



import dataclasses
import typing
from collections.abc import Sequence

from debian.deb822 import PkgRelation

from debputy.manifest_conditions import ManifestCondition
from debputy.manifest_parser.base_types import FileSystemMode
from debputy.manifest_parser.util import AttributePath


@dataclasses.dataclass(slots=True)
class DebputyCapability:
    capabilities: str
    capability_mode: FileSystemMode
    definition_source: str


class MatchedBuiltUsingRelation(typing.NamedTuple):
    is_first: bool
    relationship: "PkgRelation.ParsedRelation"


class BuiltUsingItem(typing.NamedTuple):
    matched_packages: Sequence[MatchedBuiltUsingRelation]
    conditional: ManifestCondition | None
    attribute_path: AttributePath


class StaticBuiltUsing(tuple[BuiltUsingItem, ...]):
    pass


class BuiltUsing(tuple[BuiltUsingItem, ...]):
    pass



import dataclasses
import importlib.resources
import re
from functools import lru_cache
from typing import (
    Union,
    Optional,
    TYPE_CHECKING,
    List,
    Self,
)
from collections.abc import Sequence, Mapping

from debputy.linting.lint_util import LintState, with_range_in_continuous_parts
from debputy.lsp.debputy_ls import DebputyLanguageServer
from debputy.lsp.lsp_debian_control_reference_data import (
    DebianWatch5FileMetadata,
    Deb822KnownField,
)

import debputy.lsp.data.deb822_data as deb822_ref_data_dir
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_completer,
    lsp_hover,
    lsp_standard_handler,
    lsp_folding_ranges,
    lsp_semantic_tokens_full,
    lsp_will_save_wait_until,
    lsp_format_document,
    SecondaryLanguage,
    LanguageDispatchRule,
    lsp_cli_reformat_document,
)
from debputy.lsp.lsp_generic_deb822 import (
    deb822_completer,
    deb822_hover,
    deb822_folding_ranges,
    deb822_semantic_tokens_full,
    deb822_format_file,
    scan_for_syntax_errors_and_token_level_diagnostics,
)
from debputy.lsp.lsp_reference_keyword import LSP_DATA_DOMAIN
from debputy.lsp.ref_models.deb822_reference_parse_models import (
    GenericVariable,
    GENERIC_VARIABLE_REFERENCE_DATA_PARSER,
)
from debputy.lsp.text_util import markdown_urlify
from debian._deb822_repro import (
    Deb822ParagraphElement,
)
from debputy.lsprotocol.types import (
    CompletionItem,
    CompletionList,
    CompletionParams,
    HoverParams,
    Hover,
    TEXT_DOCUMENT_CODE_ACTION,
    SemanticTokens,
    SemanticTokensParams,
    FoldingRangeParams,
    FoldingRange,
    WillSaveTextDocumentParams,
    TextEdit,
    DocumentFormattingParams,
)
from debputy.manifest_parser.util import AttributePath
from debputy.yaml import MANIFEST_YAML

try:
    from debian._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )

    from pygls.server import LanguageServer
    from pygls.workspace import TextDocument
except ImportError:
    pass


if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


_CONTAINS_SPACE_OR_COLON = re.compile(r"[\s:]")

_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debian/watch",
    None,
    "debian/watch",
    [
        # Presumably, emacs's name
        SecondaryLanguage("debian-watch"),
        # Presumably, vim's name
        SecondaryLanguage("debwatch"),
    ],
)

_DWATCH_FILE_METADATA = DebianWatch5FileMetadata()

lsp_standard_handler(_DISPATCH_RULE, TEXT_DOCUMENT_CODE_ACTION)


@dataclasses.dataclass(slots=True, frozen=True)
class VariableMetadata:
    name: str
    doc_uris: Sequence[str]
    synopsis: str
    description: str

    def render_metadata_fields(self) -> str:
        doc_uris = self.doc_uris
        parts = []
        if doc_uris:
            if len(doc_uris) == 1:
                parts.append(f"Documentation: {markdown_urlify(doc_uris[0])}")
            else:
                parts.append("Documentation:")
                parts.extend(f" - {markdown_urlify(uri)}" for uri in doc_uris)
        return "\n".join(parts)

    @classmethod
    def from_ref_data(cls, x: GenericVariable) -> "Self":
        doc = x.get("documentation", {})
        return cls(
            x["name"],
            doc.get("uris", []),
            doc.get("synopsis", ""),
            doc.get("long_description", ""),
        )


def dwatch_variables_metadata_basename() -> str:
    return "debian_watch_variables_data.yaml"


def _as_variables_metadata(
    args: list[VariableMetadata],
) -> Mapping[str, VariableMetadata]:
    r = {s.name: s for s in args}
    assert len(r) == len(args)
    return r


@lru_cache
def dwatch_variables_metadata() -> Mapping[str, VariableMetadata]:
    p = importlib.resources.files(deb822_ref_data_dir.__name__).joinpath(
        dwatch_variables_metadata_basename()
    )

    with p.open("r", encoding="utf-8") as fd:
        raw = MANIFEST_YAML.load(fd)

    attr_path = AttributePath.root_path(p)
    ref = GENERIC_VARIABLE_REFERENCE_DATA_PARSER.parse_input(raw, attr_path)
    return _as_variables_metadata(
        [VariableMetadata.from_ref_data(x) for x in ref["variables"]]
    )


def _custom_hover(
    ls: "DebputyLanguageServer",
    server_position: types.Position,
    _current_field: str | None,
    _word_at_position: str,
    _known_field: Deb822KnownField | None,
    in_value: bool,
    _doc: "TextDocument",
    lines: list[str],
) -> Hover | str | None:
    if not in_value:
        return None

    line_no = server_position.line
    line = lines[line_no]
    variable_search_ref = server_position.character
    variable = ""
    try:
        # Unlike ${} substvars where the start and end uses distinct characters, we cannot
        # know for certain whether we are at the start or end of a variable when we land
        # directly on a separator.
        try:
            variable_start = line.rindex("@", 0, variable_search_ref)
        except ValueError:
            if line[variable_search_ref] != "@":
                raise
            variable_start = variable_search_ref

        variable_end = line.index("@", variable_start + 1)
        if server_position.character <= variable_end:
            variable = line[variable_start : variable_end + 1]
    except (ValueError, IndexError):
        pass

    if variable != "" and variable != "@@":
        substvar_md = dwatch_variables_metadata().get(variable)

        if substvar_md is None:
            # In case of `@PACKAGE@-lin<CURSOR>ux-@ANY_VERSION@`
            return None
        doc = ls.translation(LSP_DATA_DOMAIN).pgettext(
            f"Variable:{substvar_md.name}",
            substvar_md.description,
        )
        md_fields = "\n" + substvar_md.render_metadata_fields()
        return f"# Variable `{variable}`\n\n{doc}{md_fields}"

    return None


@lsp_hover(_DISPATCH_RULE)
def _debian_watch_hover(
    ls: "DebputyLanguageServer",
    params: HoverParams,
) -> Hover | None:
    return deb822_hover(ls, params, _DWATCH_FILE_METADATA, custom_handler=_custom_hover)


@lsp_completer(_DISPATCH_RULE)
def _debian_watch_completions(
    ls: "DebputyLanguageServer",
    params: CompletionParams,
) -> CompletionList | Sequence[CompletionItem] | None:
    return deb822_completer(ls, params, _DWATCH_FILE_METADATA)


@lsp_folding_ranges(_DISPATCH_RULE)
def _debian_watch_folding_ranges(
    ls: "DebputyLanguageServer",
    params: FoldingRangeParams,
) -> Sequence[FoldingRange] | None:
    return deb822_folding_ranges(ls, params, _DWATCH_FILE_METADATA)


@lint_diagnostics(_DISPATCH_RULE)
async def _lint_debian_watch(lint_state: LintState) -> None:
    deb822_file = lint_state.parsed_deb822_file_content

    if not _DWATCH_FILE_METADATA.file_metadata_applies_to_file(deb822_file):
        return

    first_error = await scan_for_syntax_errors_and_token_level_diagnostics(
        deb822_file,
        lint_state,
    )
    header_stanza, source_stanza = _DWATCH_FILE_METADATA.stanza_types()
    stanza_no = 0

    async for stanza_range, stanza in lint_state.slow_iter(
        with_range_in_continuous_parts(deb822_file.iter_parts())
    ):
        if not isinstance(stanza, Deb822ParagraphElement):
            continue
        stanza_position = stanza_range.start_pos
        if stanza_position.line_position >= first_error:
            break
        stanza_no += 1
        is_source_stanza = stanza_no != 1
        if is_source_stanza:
            stanza_metadata = _DWATCH_FILE_METADATA.classify_stanza(
                stanza,
                stanza_no,
            )
            other_stanza_metadata = header_stanza
            other_stanza_name = "Header"
        elif "Version" in stanza:
            stanza_metadata = header_stanza
            other_stanza_metadata = source_stanza
            other_stanza_name = "Source"
        else:
            break

        await stanza_metadata.stanza_diagnostics(
            deb822_file,
            stanza,
            stanza_position,
            lint_state,
            confusable_with_stanza_name=other_stanza_name,
            confusable_with_stanza_metadata=other_stanza_metadata,
        )


@lsp_will_save_wait_until(_DISPATCH_RULE)
def _debian_watch_on_save_formatting(
    ls: "DebputyLanguageServer",
    params: WillSaveTextDocumentParams,
) -> Sequence[TextEdit] | None:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lint_state = ls.lint_state(doc)
    return deb822_format_file(lint_state, _DWATCH_FILE_METADATA)


@lsp_cli_reformat_document(_DISPATCH_RULE)
def _reformat_debian_watch(
    lint_state: LintState,
) -> Sequence[TextEdit] | None:
    return deb822_format_file(lint_state, _DWATCH_FILE_METADATA)


@lsp_format_document(_DISPATCH_RULE)
def _debian_watch_format_doc(
    ls: "DebputyLanguageServer",
    params: DocumentFormattingParams,
) -> Sequence[TextEdit] | None:
    doc = ls.workspace.get_text_document(params.text_document.uri)
    lint_state = ls.lint_state(doc)
    return deb822_format_file(lint_state, _DWATCH_FILE_METADATA)


@lsp_semantic_tokens_full(_DISPATCH_RULE)
async def _debian_watch_semantic_tokens_full(
    ls: "DebputyLanguageServer",
    request: SemanticTokensParams,
) -> SemanticTokens | None:
    return await deb822_semantic_tokens_full(
        ls,
        request,
        _DWATCH_FILE_METADATA,
    )



import asyncio
import collections
from typing import (
    Optional,
    Any,
    TYPE_CHECKING,
    cast,
    Union,
)
from collections.abc import Mapping, Callable, Sequence, Awaitable

import pytest

from debputy.filesystem_scan import VirtualPathBase
from debputy.linting.lint_util import (
    LinterPositionCodec,
    LintStateImpl,
    LintState,
    AsyncLinterImpl,
)
from debputy.lsp.config.config_options import (
    ALL_DEBPUTY_CONFIG_OPTIONS,
    DebputyConfigOption,
)
from debputy.lsp.config.debputy_config import DebputyConfig
from debputy.lsp.diagnostics import DiagnosticData
from debputy.lsp.maint_prefs import (
    MaintainerPreferenceTable,
    EffectiveFormattingPreference,
    determine_effective_preference,
)
from debputy.lsp.quickfixes import CODE_ACTION_HANDLERS, CodeActionName
from debputy.lsp.text_edit import apply_text_edits
from debputy.packages import DctrlParser
from debputy.plugin.api.feature_set import PluginProvidedFeatureSet
from debputy.util import T

if TYPE_CHECKING:
    from lsprotocol import types
else:
    from debputy.lsprotocol import types


try:
    from Levenshtein import distance

    HAS_LEVENSHTEIN = True
except ImportError:
    HAS_LEVENSHTEIN = False

LINTER_POSITION_CODEC = LinterPositionCodec()


class TestDebputyConfig(DebputyConfig):

    def set_config_value(
        self, config_option: DebputyConfigOption[T], value: T | None
    ) -> None:
        self._config[config_option] = value

    def reset_config(self, config_option: DebputyConfigOption[T]) -> None:
        self.set_config_value(config_option, config_option.default_value)


class BaseLintStateWrapper:
    def __init__(
        self,
        path: str,
        debputy_plugin_feature_set: PluginProvidedFeatureSet,
        dctrl_parser: DctrlParser,
    ) -> None:
        self._debputy_plugin_feature_set = debputy_plugin_feature_set
        self.dctrl_lines: list[str] | None = None
        self.path = path
        self._dctrl_parser = dctrl_parser
        self.source_root: VirtualPathBase | None = None
        self.maint_preference_table = MaintainerPreferenceTable({}, {})
        self.effective_preference: EffectiveFormattingPreference | None = None
        self.debputy_config = TestDebputyConfig(
            {c: c.default_value for c in ALL_DEBPUTY_CONFIG_OPTIONS}
        )

    def _setup(
        self,
        lines: list[str],
        *,
        lint_handler: Optional["AsyncLinterImpl"] = None,
    ) -> LintStateImpl:
        source_package = None
        binary_packages = None
        dctrl_lines = self.dctrl_lines
        if dctrl_lines is not None:
            _, source_package, binary_packages = (
                self._dctrl_parser.parse_source_debian_control(
                    dctrl_lines, ignore_errors=True
                )
            )
        source_root = self.source_root
        debian_dir = source_root.get("debian") if source_root is not None else None
        return LintStateImpl(
            self._debputy_plugin_feature_set,
            self.maint_preference_table,
            source_root,
            debian_dir,
            self.path,
            "".join(lines) if lines is not None else "",
            lines,
            self.debputy_config,
            source_package,
            binary_packages,
            self.effective_preference,
            lint_handler,
        )


class LintWrapper(BaseLintStateWrapper):

    def __init__(
        self,
        path: str,
        handler: Callable[[LintState], Awaitable[None]],
        debputy_plugin_feature_set: PluginProvidedFeatureSet,
        dctrl_parser: DctrlParser,
    ) -> None:
        super().__init__(path, debputy_plugin_feature_set, dctrl_parser)
        self._handler = handler
        self._lint_state: LintStateImpl | None = None

    @property
    def lint_state(self) -> LintStateImpl:
        lint_state = self._lint_state
        assert (
            lint_state is not None
        ), "lint_state must be fetched after the diagnostics have been run"
        return lint_state

    def __call__(self, lines: list[str]) -> list["types.Diagnostic"]:
        self._lint_state = lint_state = self._setup(lines, lint_handler=self._handler)
        return check_diagnostics(asyncio.run(lint_state.gather_diagnostics()))


class ReformatWrapper(BaseLintStateWrapper):

    def __init__(
        self,
        path: str,
        handler: Callable[[LintState], Sequence[types.TextEdit] | None],
        debputy_plugin_feature_set: PluginProvidedFeatureSet,
        dctrl_parser: DctrlParser,
        maint_preference_table: MaintainerPreferenceTable,
    ) -> None:
        super().__init__(path, debputy_plugin_feature_set, dctrl_parser)
        self._handler = handler
        self.maint_preference_table = maint_preference_table

    def reformat(self, lines: list[str]) -> list["types.TextEdit"]:
        def _style(lint_state: LintState) -> EffectiveFormattingPreference | None:
            pref_result = determine_effective_preference(
                lint_state.maint_preference_table, lint_state.source_package, None
            )
            return pref_result[0] if pref_result else None

        return self._reformat(lines, _style)

    def reformat_with_named_style(
        self, named_style: str, lines: list[str]
    ) -> list["types.TextEdit"]:
        return self._reformat(
            lines, lambda s: s.maint_preference_table.named_styles[named_style]
        )

    def _reformat(
        self,
        lines: list[str],
        style_selector: Callable[[LintState], EffectiveFormattingPreference | None],
    ) -> list["types.TextEdit"]:
        lint_state = self._setup(lines, lint_handler=None)
        lint_state.effective_preference = style_selector(lint_state)
        reformat_result = self._handler(lint_state)
        return reformat_result if reformat_result is not None else []


def requires_levenshtein(func: Any) -> Any:
    return pytest.mark.skipif(
        not HAS_LEVENSHTEIN, reason="Missing python3-levenshtein"
    )(func)


def check_diagnostics(
    diagnostics: list["types.Diagnostic"] | None,
) -> list["types.Diagnostic"]:
    if diagnostics:
        for diagnostic in diagnostics:
            assert diagnostic.severity is not None
            assert diagnostic.source is not None
    elif diagnostics is None:
        diagnostics = []
    return diagnostics


def by_range_sort_key(diagnostic: types.Diagnostic) -> Any:
    start_pos = diagnostic.range.start
    end_pos = diagnostic.range.end
    return start_pos.line, start_pos.character, end_pos.line, end_pos.character


def group_diagnostics_by_severity(
    diagnostics: list["types.Diagnostic"] | None,
) -> Mapping["types.DiagnosticSeverity", list["types.Diagnostic"]]:
    if not diagnostics:
        return {}

    by_severity = collections.defaultdict(list)

    for diagnostic in sorted(diagnostics, key=by_range_sort_key):
        severity = diagnostic.severity
        assert severity is not None
        by_severity[severity].append(diagnostic)

    return by_severity


def diag_range_to_text(lines: Sequence[str], range_: "types.Range") -> str:
    parts = []
    for line_no in range(range_.start.line, range_.end.line + 1):
        line = lines[line_no]
        chunk = line
        if line_no == range_.start.line and line_no == range_.end.line:
            chunk = line[range_.start.character : range_.end.character]
        elif line_no == range_.start.line:
            chunk = line[range_.start.character :]
        elif line_no == range_.end.line:
            chunk = line[: range_.end.character]
        parts.append(chunk)
    return "".join(parts)


def lsp_range_to_text(lines: Sequence[str], range_: "types.Range") -> str:
    return diag_range_to_text(lines, range_)


def standard_quick_fixes_for_diagnostic(
    lint_state: LintState,
    code_action_params: "types.CodeActionParams",
    diagnostic: "types.Diagnostic",
) -> Sequence[Union["types.CodeAction", "types.Command"]]:
    if not isinstance(diagnostic.data, dict):
        return []
    data: DiagnosticData = cast("DiagnosticData", diagnostic.data)
    quickfixes = data.get("quickfixes")
    if quickfixes is None:
        return []
    actions = []
    for action_suggestion in quickfixes:
        if (
            action_suggestion
            and isinstance(action_suggestion, Mapping)
            and "code_action" in action_suggestion
        ):
            action_name: CodeActionName = action_suggestion["code_action"]
            handler = CODE_ACTION_HANDLERS.get(action_name)
            if handler is not None:
                actions.extend(
                    handler(
                        lint_state,
                        cast("Mapping[str, str]", action_suggestion),
                        code_action_params,
                        diagnostic,
                    )
                )
            else:
                assert False, f"No codeAction handler for {action_name} !?"
    return actions


def apply_code_action_edits(
    doc_uri: str, lines: list[str], code_action: types.CodeAction
) -> str:
    """Apply the text edits for a given doc URI from a code action


    The code action must have at least one edit for the given document. It *may*
    have edits for other documents and these will be ignored (except when they
    use an unsupported feature).

    Unsupported features:
     * Any code action with a command
     * Any code action with non-text related edits (create file, rename file, etc.)

    :param doc_uri: The doc_uri for the document represented by the lines parameter
    :param lines: The lines of the document to apply the edit to
    :param code_action: The code action to apply
    :return: The file after edits.
    """
    workspace_edit = code_action.edit
    assert workspace_edit is not None
    # We cannot apply a command with this function
    assert code_action.command is None
    orig = "".join(lines)
    document_changes = workspace_edit.document_changes
    if document_changes:
        assert workspace_edit.changes is None
        # We cannot apply non-edits
        assert all(isinstance(e, types.TextDocumentEdit) for e in document_changes)
        changes = cast("Sequence[types.TextDocumentEdit", document_changes)
        doc_edits = [c.edits for c in changes if c.text_document.uri == doc_uri]
        assert len(doc_edits) == 1
        edits = doc_edits[0]
    else:
        changes = workspace_edit.changes
        assert changes is not None
        edits = changes.get(doc_uri)
        assert edits
    return apply_text_edits(orig, lines, edits)


def apply_formatting_edits(
    lines: list[str],
    edits: Sequence[types.TextEdit],
) -> str:
    orig = "".join(lines)
    return apply_text_edits(orig, lines, edits)


def code_actions_for(
    uri: str,
    text_range: "types.Range",
) -> "types.CodeActionsParams":
    return types.CodeActionParams(
        types.TextDocumentIdentifier(uri),
        text_range,
        # Ignore the actual list of diagnostics, since we do not use them.
        types.CodeActionContext([]),
    )



from collections.abc import Mapping
from debputy.plugin.api.impl_types import (
    KNOWN_PACKAGING_FILE_CATEGORY_DESCRIPTIONS,
    KNOWN_PACKAGING_FILE_CONFIG_FEATURE_DESCRIPTION,
    KnownPackagingConfigFeature,
    KnownPackagingFileCategories,
    ReferenceValue,
)

REFERENCE_DATA_TABLE: Mapping[
    str,
    Mapping[KnownPackagingConfigFeature, ReferenceValue]
    | Mapping[KnownPackagingFileCategories, ReferenceValue],
] = {
    "config-features": KNOWN_PACKAGING_FILE_CONFIG_FEATURE_DESCRIPTION,
    "file-categories": KNOWN_PACKAGING_FILE_CATEGORY_DESCRIPTIONS,
}



import dataclasses
from typing import TypedDict, TYPE_CHECKING, Generic, Optional
from collections.abc import Callable

from debputy.plugin.plugin_state import current_debputy_plugin_required

if TYPE_CHECKING:
    from debputy.manifest_parser.parser_data import ParserContextData
    from debputy.manifest_parser.util import AttributePath


class DebputyParsedContent(TypedDict):
    pass


class DebputyDispatchableType:
    __slots__ = ("_debputy_plugin",)

    def __init__(self) -> None:
        self._debputy_plugin = current_debputy_plugin_required()


@dataclasses.dataclass
class TypeMapping[S: (int, str), T]:
    target_type: type[T]
    source_type: type[S]
    mapper: Callable[[S, "AttributePath", Optional["ParserContextData"]], T]



import dataclasses
import os
import re
import textwrap
import typing
from typing import (
    Any,
    Iterator,
    NotRequired,
    Union,
    Literal,
    TypedDict,
    Annotated,
    Self,
    cast,
)

from debian.deb822 import PkgRelation

from debputy._manifest_constants import (
    MK_INSTALLATION_SEARCH_DIRS,
    MK_BINARY_VERSION,
    MK_SERVICES,
)
from debputy.maintscript_snippet import (
    DpkgMaintscriptHelperCommand,
    MaintscriptSnippet,
    SnippetResolver,
)
from debputy.manifest_parser.base_types import (
    DebputyParsedContentStandardConditional,
    FileSystemExactMatchRule,
)
from debputy.manifest_parser.declarative_parser import ParserGenerator
from debputy.manifest_parser.exceptions import ManifestParseException
from debputy.manifest_parser.parse_hints import DebputyParseHint
from debputy.manifest_parser.parser_data import ParserContextData
from debputy.manifest_parser.tagging_types import DebputyParsedContent
from debputy.manifest_parser.util import AttributePath
from debputy.path_matcher import MatchRule, MATCH_ANYTHING, ExactFileSystemPath
from debputy.plugin.api import reference_documentation
from debputy.plugin.api.impl import (
    DebputyPluginInitializerProvider,
    ServiceDefinitionImpl,
)
from debputy.plugin.api.parser_tables import OPARSER_PACKAGES
from debputy.plugin.api.spec import (
    ServiceUpgradeRule,
    ServiceDefinition,
    DSD,
    INTEGRATION_MODE_DH_DEBPUTY_RRR,
    not_integrations,
    documented_attr,
)
from debputy.plugins.debputy.types import (
    BuiltUsingItem,
    BuiltUsing,
    StaticBuiltUsing,
    MatchedBuiltUsingRelation,
)
from debputy.transformation_rules import TransformationRule
from debputy.util import _error, manifest_format_doc

ACCEPTABLE_CLEAN_ON_REMOVAL_FOR_GLOBS_AND_EXACT_MATCHES = frozenset(
    [
        "./var/log",
    ]
)


ACCEPTABLE_CLEAN_ON_REMOVAL_IF_EXACT_MATCH_OR_SUBDIR_OF = frozenset(
    [
        "./etc",
        "./run",
        "./var/lib",
        "./var/cache",
        "./var/backups",
        "./var/spool",
        # linux-image uses these paths with some `rm -f`
        "./usr/lib/modules",
        "./lib/modules",
        # udev special case
        "./lib/udev",
        "./usr/lib/udev",
        # pciutils deletes /usr/share/misc/pci.ids.<ext>
        "./usr/share/misc",
    ]
)


def register_binary_package_rules(api: DebputyPluginInitializerProvider) -> None:
    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        MK_BINARY_VERSION,
        BinaryVersionParsedFormat,
        _parse_binary_version,
        source_format=str,
        register_value=False,
    )

    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        "transformations",
        list[TransformationRule],
        _unpack_list,
        register_value=False,
    )

    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        "conffile-management",
        list[DpkgMaintscriptHelperCommand],
        _unpack_list,
        expected_debputy_integration_mode=not_integrations(
            INTEGRATION_MODE_DH_DEBPUTY_RRR
        ),
        register_value=False,
    )

    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        MK_SERVICES,
        list[ServiceRuleParsedFormat],
        _process_service_rules,
        source_format=list[ServiceRuleSourceFormat],
        expected_debputy_integration_mode=not_integrations(
            INTEGRATION_MODE_DH_DEBPUTY_RRR
        ),
        register_value=False,
    )

    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        "dpkg-gensymbols",
        DpkgGensymbolsOptionsFormat,
        DpkgGensymbolsOptions.parse,
        expected_debputy_integration_mode=not_integrations(
            INTEGRATION_MODE_DH_DEBPUTY_RRR
        ),
        # Pulled by makeshlibs.py, so registration is needed.
        register_value=True,
        inline_reference_documentation=reference_documentation(
            title="Configure `dpkg-gensmybols` options (`$RULE_NAME`)",
            description=textwrap.dedent(
                """\
                Configure the `dpkg-gensymbols` for the given binary
                package.

                Examples:

                    packages:
                      PKG:
                        $RULE_NAME:
                          # Pass `-c4` to `dpkg-gensymbols`
                          check-level: 4
                """,
            ),
            attributes=[
                documented_attr(
                    "check_level",
                    textwrap.dedent("""\
                        Configure the check-level (`-c`) for `dpkg-gensymbols`

                        The levels are defined in [man:dpkg-gensymbols.1] as:

                         * `0`: Never fails.
                         * `1`: Fails if some symbols have disappeared.
                         * `2`: Fails if some new symbols have been introduced.
                         * `3`: Fails if some libraries have disappeared.
                         * `4`: Fails if some libraries have been introduced.

                        The higher levels include all the checks from lower levels.
                        As an example, using a value of `3` would include all the
                        checks from `3`, `2`, and `1` at the same time.

                        The default check-level in `dpkg-gensymbols` is `1`.

                        [man:dpkg-gensymbols.1]: https://manpages.debian.org/dpkg-gensymbols.1
                    """),
                ),
            ],
        ),
    )

    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        "clean-after-removal",
        ListParsedFormat,
        _parse_clean_after_removal,
        # FIXME: debputy won't see the attributes for this one :'(
        # (update `debputy_docs.yaml` when fixed)
        source_format=list[Any],
        expected_debputy_integration_mode=not_integrations(
            INTEGRATION_MODE_DH_DEBPUTY_RRR
        ),
        register_value=False,
    )

    api.pluggable_manifest_rule(
        OPARSER_PACKAGES,
        MK_INSTALLATION_SEARCH_DIRS,
        InstallationSearchDirsParsedFormat,
        _parse_installation_search_dirs,
        source_format=list[FileSystemExactMatchRule],
        expected_debputy_integration_mode=not_integrations(
            INTEGRATION_MODE_DH_DEBPUTY_RRR
        ),
        register_value=False,
    )

    api.pluggable_manifest_rule(
        rule_type=OPARSER_PACKAGES,
        rule_name="built-using",
        parsed_format=list[BuiltUsingParsedFormat],
        handler=_parse_built_using,
        expected_debputy_integration_mode=not_integrations(
            "dh-sequence-zz-debputy-rrr"
        ),
        inline_reference_documentation=reference_documentation(
            title="Built-Using dependency relations (`$RULE_NAME`)",
            description=textwrap.dedent(
                """\
                Generate a `Built-Using` dependency relation on the
                build dependencies selected by the `sources-for`, which
                may contain a `*` wildcard matching any number of
                arbitrary characters.

                The `built-using` should be used for static linking
                where license of dependency libraries require the
                exact source to be retained. Usually these libraries
                will be under the license terms like GNU GPL.

                    packages:
                      PKG:
                        $RULE_NAME:
                        - sources-for: foo-*-source   # foo-3.1.0-source
                        - sources-for: librust-*-dev  # several matches
                        - sources-for: foo
                          when:                       # foo is always installed
                            arch-matches: amd64       # but only used on amd64

                Either of these conditions prevents the generation:
                * PKG is not part of the current build because of its
                  `Architecture` or `Build-Profiles` fields.
                * The match in `Build-Depends` carries an
                  architecture or build profile restriction that does
                  not match the current run.
                * The match in `Build-Depends` is not installed.
                  This should only happen inside alternatives, see below.
                * The manifest item carries a `when:` condition that
                  evaluates to false. This may be useful when the match
                  must be installed for unrelated reasons.

                Matches are searched in the `Build-Depends` field of
                the source package, and either `Build-Depends-Indep`
                or `Build-Depends-Arch` depending on PKG.

                In alternatives like `a | b`, each option may match
                separately.  This is a compromise between
                reproducibility on automatic builders (where the set
                of installed package is constant), and least surprise
                during local builds (where `b` may be installed
                alone). There seems to be no one-size fits all
                solution when both are installed.

                Architecture qualifiers and version restrictions in
                `Build-Depends` are ignored. The only allowed
                co-installations require a common source and version.
                """,
            ),
        ),
    )

    api.pluggable_manifest_rule(
        rule_type=OPARSER_PACKAGES,
        rule_name="static-built-using",
        parsed_format=list[BuiltUsingParsedFormat],
        handler=_parse_static_built_using,
        expected_debputy_integration_mode=not_integrations(
            "dh-sequence-zz-debputy-rrr"
        ),
        inline_reference_documentation=reference_documentation(
            title="Static-Built-Using dependency relations (`$RULE_NAME`)",
            description=textwrap.dedent(
                """\
                Generate a `Static-Built-Using` dependency relation on the
                build dependencies selected by the `sources-for`, which
                may contain a `*` wildcard matching any number of
                arbitrary characters.

                The `static-built-using` should be used for static linking
                where license of dependency libraries do not require the
                exact source to be retained. This is usually libraries under
                permissive libraries like Apache-2.0 or MIT/X11/Expat.

                    packages:
                      PKG:
                        $RULE_NAME:
                        - sources-for: foo-*-source   # foo-3.1.0-source
                        - sources-for: librust-*-dev  # several matches
                        - sources-for: foo
                          when:                       # foo is always installed
                            arch-matches: amd64       # but only used on amd64

                Either of these conditions prevents the generation:
                * PKG is not part of the current build because of its
                  `Architecture` or `Build-Profiles` fields.
                * The match in `Build-Depends` carries an
                  architecture or build profile restriction that does
                  not match the current run.
                * The match in `Build-Depends` is not installed.
                  This should only happen inside alternatives, see below.
                * The manifest item carries a `when:` condition that
                  evaluates to false. This may be useful when the match
                  must be installed for unrelated reasons.

                Matches are searched in the `Build-Depends` field of
                the source package, and either `Build-Depends-Indep`
                or `Build-Depends-Arch` depending on PKG.

                In alternatives like `a | b`, each option may match
                separately.  This is a compromise between
                reproducibility on automatic builders (where the set
                of installed package is constant), and least surprise
                during local builds (where `b` may be installed
                alone). There seems to be no one-size fits all
                solution when both are installed.

                Architecture qualifiers and version restrictions in
                `Build-Depends` are ignored. The only allowed
                co-installations require a common source and version.
                """,
            ),
        ),
    )


class ServiceRuleSourceFormat(TypedDict):
    service: str
    type_of_service: NotRequired[str]
    service_scope: NotRequired[Literal["system", "user"]]
    enable_on_install: NotRequired[bool]
    start_on_install: NotRequired[bool]
    on_upgrade: NotRequired[ServiceUpgradeRule]
    service_manager: NotRequired[
        Annotated[str, DebputyParseHint.target_attribute("service_managers")]
    ]
    service_managers: NotRequired[list[str]]


class ServiceRuleParsedFormat(DebputyParsedContent):
    service: str
    type_of_service: NotRequired[str]
    service_scope: NotRequired[Literal["system", "user"]]
    enable_on_install: NotRequired[bool]
    start_on_install: NotRequired[bool]
    on_upgrade: NotRequired[ServiceUpgradeRule]
    service_managers: NotRequired[list[str]]


@dataclasses.dataclass(slots=True, frozen=True)
class ServiceRule:
    definition_source: str
    service: str
    type_of_service: str
    service_scope: Literal["system", "user"]
    enable_on_install: bool | None
    start_on_install: bool | None
    on_upgrade: ServiceUpgradeRule | None
    service_managers: frozenset[str] | None

    @classmethod
    def from_service_rule_parsed_format(
        cls,
        data: ServiceRuleParsedFormat,
        attribute_path: AttributePath,
    ) -> "Self":
        service_managers = data.get("service_managers")
        return cls(
            attribute_path.path,
            data["service"],
            data.get("type_of_service", "service"),
            cast("Literal['system', 'user']", data.get("service_scope", "system")),
            data.get("enable_on_install"),
            data.get("start_on_install"),
            data.get("on_upgrade"),
            None if service_managers is None else frozenset(service_managers),
        )

    def applies_to_service_manager(self, service_manager: str) -> bool:
        return self.service_managers is None or service_manager in self.service_managers

    def apply_to_service_definition(
        self,
        service_definition: ServiceDefinition[DSD],
    ) -> ServiceDefinition[DSD]:
        assert isinstance(service_definition, ServiceDefinitionImpl)
        if not service_definition.is_plugin_provided_definition:
            _error(
                f"Conflicting definitions related to {self.service} (type: {self.type_of_service},"
                f" scope: {self.service_scope}). First definition at {service_definition.definition_source},"
                f" the second at {self.definition_source}). If they are for different service managers,"
                " you can often avoid this problem by explicitly defining which service managers are applicable"
                ' to each rule via the "service-managers" keyword.'
            )
        changes = {
            "definition_source": self.definition_source,
            "is_plugin_provided_definition": False,
        }
        if (
            self.service != service_definition.name
            and self.service in service_definition.names
        ):
            changes["name"] = self.service
        if self.enable_on_install is not None:
            changes["auto_start_on_install"] = self.enable_on_install
        if self.start_on_install is not None:
            changes["auto_start_on_install"] = self.start_on_install
        if self.on_upgrade is not None:
            changes["on_upgrade"] = self.on_upgrade

        return service_definition.replace(**changes)


class BinaryVersionParsedFormat(DebputyParsedContent):
    binary_version: str


class BuiltUsingParsedFormat(DebputyParsedContentStandardConditional):
    """Also used for static-built-using."""

    sources_for: str


class ListParsedFormat(DebputyParsedContent):
    elements: list[Any]


class ListOfTransformationRulesFormat(DebputyParsedContent):
    elements: list[TransformationRule]


class ListOfDpkgMaintscriptHelperCommandFormat(DebputyParsedContent):
    elements: list[DpkgMaintscriptHelperCommand]


class InstallationSearchDirsParsedFormat(DebputyParsedContent):
    installation_search_dirs: list[FileSystemExactMatchRule]


type DpkgGensymbolsCheckLevel = typing.Literal[0, 1, 2, 3, 4]


class DpkgGensymbolsOptionsFormat(DebputyParsedContent):
    check_level: DpkgGensymbolsCheckLevel


@dataclasses.dataclass
class DpkgGensymbolsOptions:
    # We leave `check_level` optional at this stage to ensure consuming code is
    # ready for it being optional later even though it is required now.
    check_level: DpkgGensymbolsCheckLevel | None

    @classmethod
    def parse(
        cls,
        _name: str,
        parsed_data: DpkgGensymbolsOptionsFormat,
        _attribute_path: AttributePath,
        _parser_context: ParserContextData,
    ) -> typing.Self:
        return cls(**parsed_data)


def _parse_binary_version(
    _name: str,
    parsed_data: BinaryVersionParsedFormat,
    _attribute_path: AttributePath,
    _parser_context: ParserContextData,
) -> str:
    return parsed_data["binary_version"]


def _parse_installation_search_dirs(
    _name: str,
    parsed_data: InstallationSearchDirsParsedFormat,
    _attribute_path: AttributePath,
    _parser_context: ParserContextData,
) -> list[FileSystemExactMatchRule]:
    return parsed_data["installation_search_dirs"]


def _process_service_rules(
    _name: str,
    parsed_data: list[ServiceRuleParsedFormat],
    attribute_path: AttributePath,
    _parser_context: ParserContextData,
) -> list[ServiceRule]:
    return [
        ServiceRule.from_service_rule_parsed_format(x, attribute_path[i])
        for i, x in enumerate(parsed_data)
    ]


def _parse_built_using(
    _name: str,
    parsed_data: list[BuiltUsingParsedFormat],
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> BuiltUsing:
    items = _built_using_handler(parsed_data, attribute_path, parser_context)
    return BuiltUsing(items)


def _parse_static_built_using(
    _name: str,
    parsed_data: list[BuiltUsingParsedFormat],
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> StaticBuiltUsing:
    items = _built_using_handler(parsed_data, attribute_path, parser_context)
    return StaticBuiltUsing(items)


_VALID_BUILT_USING_GLOB = re.compile("[a-z*][a-z0-9.+*-]*")
_BUILT_USING_GLOB_TO_RE = str.maketrans({".": "[.]", "+": "[+]", "*": ".*"})


def _built_using_matches(
    regex: re.Pattern,
    other: Literal["Build-Depends-Arch", "Build-Depends-Indep"],
    parser_context: ParserContextData,
) -> Iterator[MatchedBuiltUsingRelation]:
    """Helper for _validate_built_using."""
    for bd_field in ("Build-Depends", other):
        raw = parser_context.source_package.fields.get(bd_field)
        if raw is not None:
            for options in PkgRelation.parse_relations(raw):
                for idx, relation in enumerate(options):
                    if regex.fullmatch(relation["name"]) is not None:
                        yield MatchedBuiltUsingRelation(not idx, relation)


def _validate_built_using(
    parsed_data: BuiltUsingParsedFormat,
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> BuiltUsingItem:
    """Helper for _built_using_handler."""
    raw_glob = parsed_data["sources_for"]
    if _VALID_BUILT_USING_GLOB.fullmatch(raw_glob) is None:
        raise ManifestParseException(
            f"The glob {raw_glob!r} defined at {attribute_path["sources_for"].path} contained invalid characters."
            f" It must only characters valid in a package name plus the `*` character"
        )
    regex = re.compile(raw_glob.translate(_BUILT_USING_GLOB_TO_RE))

    pkg = parser_context.current_binary_package_state.binary_package
    other: Literal["Build-Depends-Arch", "Build-Depends-Indep"]
    if pkg.is_arch_all:
        other = "Build-Depends-Indep"
    else:
        other = "Build-Depends-Arch"
    matched_packages = tuple(_built_using_matches(regex, other, parser_context))
    if not matched_packages:
        raise ManifestParseException(
            f"The glob {raw_glob!r} defined at {attribute_path["sources_for"].path} matches no clause of Build-Depends or {other}."
            " Either a Build-dependency is missing or the glob fails to match the intended build-dependency, or the glob superfluous and can be removed."
        )
    return BuiltUsingItem(
        matched_packages,
        parsed_data.get("when"),
        attribute_path,
    )


def _built_using_handler(
    parsed_data: list[BuiltUsingParsedFormat],
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> Iterator[BuiltUsingItem]:
    """Helper for _parse_built_using and _parse_static_built_using."""
    for idx, pd in enumerate(parsed_data):
        yield _validate_built_using(pd, attribute_path[idx], parser_context)


def _unpack_list(
    _name: str,
    parsed_data: list[Any],
    _attribute_path: AttributePath,
    _parser_context: ParserContextData,
) -> list[Any]:
    return parsed_data


class CleanAfterRemovalRuleSourceFormat(TypedDict):
    path: NotRequired[Annotated[str, DebputyParseHint.target_attribute("paths")]]
    paths: NotRequired[list[str]]
    delete_on: NotRequired[Literal["purge", "removal"]]
    recursive: NotRequired[bool]
    ignore_non_empty_dir: NotRequired[bool]


class CleanAfterRemovalRule(DebputyParsedContent):
    paths: list[str]
    delete_on: NotRequired[Literal["purge", "removal"]]
    recursive: NotRequired[bool]
    ignore_non_empty_dir: NotRequired[bool]


# FIXME: Not optimal that we are doing an initialization of ParserGenerator here. But the rule is not depending on any
#  complex types that is registered by plugins, so it will work for now.
_CLEAN_AFTER_REMOVAL_RULE_PARSER = ParserGenerator().generate_parser(
    CleanAfterRemovalRule,
    source_content=Union[CleanAfterRemovalRuleSourceFormat, str, list[str]],
    inline_reference_documentation=reference_documentation(
        reference_documentation_url=manifest_format_doc(
            "remove-runtime-created-paths-on-purge-or-post-removal-clean-after-removal"
        ),
    ),
)


# Order between clean_on_removal and conffile_management is
# important. We want the dpkg conffile management rules to happen before the
# clean clean_on_removal rules.  Since the latter only affects `postrm`
# and the order is reversed for `postrm` scripts (among other), we need do
# clean_on_removal first to account for the reversing of order.
#
# FIXME: All of this is currently not really possible todo, but it should be.
# (I think it is the correct order by "mistake" rather than by "design", which is
# what this note is about)
def _parse_clean_after_removal(
    _name: str,
    parsed_data: ListParsedFormat,
    attribute_path: AttributePath,
    parser_context: ParserContextData,
) -> None:  # TODO: Return and pass to a maintscript helper
    raw_clean_after_removal = parsed_data["elements"]
    package_state = parser_context.current_binary_package_state

    for no, raw_transformation in enumerate(raw_clean_after_removal):
        definition_source = attribute_path[no]
        clean_after_removal_rules = _CLEAN_AFTER_REMOVAL_RULE_PARSER.parse_input(
            raw_transformation,
            definition_source,
            parser_context=parser_context,
        )
        patterns = clean_after_removal_rules["paths"]
        if patterns:
            definition_source.path_hint = patterns[0]
        delete_on = clean_after_removal_rules.get("delete_on") or "purge"
        recurse = clean_after_removal_rules.get("recursive") or False
        ignore_non_empty_dir = (
            clean_after_removal_rules.get("ignore_non_empty_dir") or False
        )
        if delete_on == "purge":
            condition = '[ "$1" = "purge" ]'
        else:
            condition = '[ "$1" = "remove" ]'

        if ignore_non_empty_dir:
            if recurse:
                raise ManifestParseException(
                    'The "recursive" and "ignore-non-empty-dir" options are mutually exclusive.'
                    f" Both were enabled at the same time in at {definition_source.path}"
                )
            for pattern in patterns:
                if not pattern.endswith("/"):
                    raise ManifestParseException(
                        'When ignore-non-empty-dir is True, then all patterns must end with a literal "/"'
                        f' to ensure they only apply to directories. The pattern "{pattern}" at'
                        f" {definition_source.path} did not."
                    )

        substitution = parser_context.substitution
        match_rules = [
            MatchRule.from_path_or_glob(
                p, definition_source.path, substitution=substitution
            )
            for p in patterns
        ]
        content_lines = [
            f"if {condition}; then\n",
        ]
        for idx, match_rule in enumerate(match_rules):
            original_pattern = patterns[idx]
            if match_rule is MATCH_ANYTHING:
                raise ManifestParseException(
                    f'Using "{original_pattern}" in a clean rule would trash the system.'
                    f" Please restrict this pattern at {definition_source.path} considerably."
                )
            is_subdir_match = False
            matched_directory: str | None
            if isinstance(match_rule, ExactFileSystemPath):
                matched_directory = (
                    os.path.dirname(match_rule.path)
                    if match_rule.path not in ("/", ".", "./")
                    else match_rule.path
                )
                is_subdir_match = True
            else:
                matched_directory = getattr(match_rule, "directory", None)

            if matched_directory is None:
                raise ManifestParseException(
                    f'The pattern "{original_pattern}" defined at {definition_source.path} is not'
                    f" trivially anchored in a specific directory. Cowardly refusing to use it"
                    f" in a clean rule as it may trash the system if the pattern is overreaching."
                    f" Please avoid glob characters in the top level directories."
                )
            assert matched_directory.startswith("./") or matched_directory in (
                ".",
                "./",
                "",
            )
            acceptable_directory = False
            would_have_allowed_direct_match = False
            while matched_directory not in (".", "./", ""):
                # Our acceptable paths set includes "/var/lib" or "/etc".  We require that the
                # pattern is either an exact match, in which case it may match directly inside
                # the acceptable directory OR it is a pattern against a subdirectory of the
                # acceptable path. As an example:
                #
                # /etc/inputrc <-- OK, exact match
                # /etc/foo/*   <-- OK, subdir match
                # /etc/*       <-- ERROR, glob directly in the accepted directory.
                if is_subdir_match and (
                    matched_directory
                    in ACCEPTABLE_CLEAN_ON_REMOVAL_IF_EXACT_MATCH_OR_SUBDIR_OF
                ):
                    acceptable_directory = True
                    break
                if (
                    matched_directory
                    in ACCEPTABLE_CLEAN_ON_REMOVAL_FOR_GLOBS_AND_EXACT_MATCHES
                ):
                    # Special-case: In some directories (such as /var/log), we allow globs directly.
                    # Notably, X11's log files are /var/log/Xorg.*.log
                    acceptable_directory = True
                    break
                if (
                    matched_directory
                    in ACCEPTABLE_CLEAN_ON_REMOVAL_IF_EXACT_MATCH_OR_SUBDIR_OF
                ):
                    would_have_allowed_direct_match = True
                    break
                matched_directory = os.path.dirname(matched_directory)
                is_subdir_match = True

            if would_have_allowed_direct_match and not acceptable_directory:
                raise ManifestParseException(
                    f'The pattern "{original_pattern}" defined at {definition_source.path} seems to'
                    " be overreaching. If it had been a path (and not use a glob), the rule would"
                    " have been permitted."
                )
            elif not acceptable_directory:
                raise ManifestParseException(
                    f'The pattern or path "{original_pattern}" defined at {definition_source.path} seems to'
                    f' be overreaching or not limited to the set of "known acceptable" directories.'
                )

            try:
                shell_escaped_pattern = match_rule.shell_escape_pattern()
            except TypeError:
                raise ManifestParseException(
                    f'Sorry, the pattern "{original_pattern}" defined at {definition_source.path}'
                    f" is unfortunately not supported by `debputy` for clean-after-removal rules."
                    f" If you can rewrite the rule to something like `/var/log/foo/*.log` or"
                    f' similar "trivial" patterns. You may have to rewrite the pattern the rule '
                    f" into multiple patterns to achieve this.  This restriction is to enable "
                    f' `debputy` to ensure the pattern is correctly executed plus catch "obvious'
                    f' system trashing" patterns. Apologies for the inconvenience.'
                )

            if ignore_non_empty_dir:
                cmd = f'    rmdir --ignore-fail-on-non-empty "${{DPKG_ROOT}}"{shell_escaped_pattern}\n'
            elif recurse:
                cmd = f'    rm -fr "${{DPKG_ROOT}}"{shell_escaped_pattern}\n'
            elif original_pattern.endswith("/"):
                cmd = f'    rmdir "${{DPKG_ROOT}}"{shell_escaped_pattern}\n'
            else:
                cmd = f'    rm -f "${{DPKG_ROOT}}"{shell_escaped_pattern}\n'
            content_lines.append(cmd)
        content_lines.append("fi\n")

        snippet = MaintscriptSnippet(
            definition_source.path,
            SnippetResolver.snippet_template("".join(content_lines)),
        )
        package_state.maintscript_snippets["postrm"].append(snippet)



from typing import Iterable

import pytest

from debputy.lsp.spellchecking import (
    _HAS_HUNSPELL,
    HunspellSpellchecker,
    _builtin_exception_words,
    _NO_CORRECTIONS,
    Spellchecker,
)


@pytest.fixture(scope="session")
def test_spellchecker() -> Spellchecker:

    class TestSpellchecker(Spellchecker):
        def provide_corrections_for(self, word: str) -> Iterable[str]:
            raise NotImplementedError

    return TestSpellchecker()


@pytest.fixture(scope="session")
def hunspell_spellchecker() -> Spellchecker:
    if not _HAS_HUNSPELL:
        pytest.skip("Missing python3-hunspell hunspell-en-us")
        return

    return HunspellSpellchecker()


@pytest.mark.skipif(not _HAS_HUNSPELL, reason="Missing python3-hunspell hunspell-en-us")
def test_hunspell_checker(hunspell_spellchecker: Spellchecker) -> None:
    r = frozenset(hunspell_spellchecker.provide_corrections_for("tets"))
    assert r
    assert "test" in r


@pytest.mark.skipif(not _HAS_HUNSPELL, reason="Missing python3-hunspell hunspell-en-us")
def test_hunspell_respects_known_exclusions(
    hunspell_spellchecker: Spellchecker,
) -> None:
    for word in _builtin_exception_words():
        assert hunspell_spellchecker.provide_corrections_for(word) is _NO_CORRECTIONS


@pytest.mark.parametrize(
    "excluded_word",
    [
        "dpkg-foo",
        "dh_foo",
        "dh-sequence-foo",
        "update-foo",
        "debconf-foo",
        "DEB_FOO",
        "DPKG_FOO",
    ],
)
@pytest.mark.skipif(not _HAS_HUNSPELL, reason="Missing python3-hunspell hunspell-en-us")
def test_hunspell_respects_word_level_exclusions(
    hunspell_spellchecker: Spellchecker, excluded_word: str
) -> None:
    assert (
        hunspell_spellchecker.provide_corrections_for(excluded_word) is _NO_CORRECTIONS
    )


def test_iter_words(test_spellchecker: Spellchecker) -> None:
    line = "  * debputy: Avoid installing unnecessary files into `udeb` packages"
    words = list(test_spellchecker.iter_words(line))

    # Validate that `udeb` is excluded and words have been isolated
    assert [w[0] for w in words] == [
        "debputy",
        "Avoid",
        "installing",
        "unnecessary",
        "files",
        "into",
        "packages",
    ]

    # Check the word positions
    for word, start, end in words:
        assert line[start:end] == word


@pytest.mark.parametrize(
    "excluded_word",
    [
        "31_aide_ldconfig",
        "SCRIPTRETVAL",
        "$SCRIPTRETVAL",
        "{script_retval}",
        "32bit",
        "<some@email.com>",
    ],
)
def test_general_exceptions(
    test_spellchecker: Spellchecker, excluded_word: str
) -> None:
    line = f"some phrase where {excluded_word} appears"
    words = list(test_spellchecker.iter_words(line))
    # Assert we are excluding the word
    assert [w[0] for w in words] == ["some", "phrase", "where", "appears"]

    for word, start, end in words:
        assert line[start:end] == word



from typing import (
    Optional,
    Union,
    TYPE_CHECKING,
)
from collections.abc import Sequence

from debputy.linting.lint_util import LintState
from debputy.lsp.config.parser import DEBPUTY_CONFIG_PARSER
from debputy.lsp.lsp_features import (
    lint_diagnostics,
    lsp_standard_handler,
    lsp_hover,
    lsp_completer,
    LanguageDispatchRule,
    SecondaryLanguage,
)
from debputy.lsp.lsp_generic_yaml import (
    generic_yaml_hover,
    LSPYAMLHelper,
    generic_yaml_lint,
    generic_yaml_completer,
)

try:
    from debputy.lsp.debputy_ls import DebputyLanguageServer
    from debian._deb822_repro.locatable import (
        Position as TEPosition,
        Range as TERange,
    )
except ImportError:
    pass

if TYPE_CHECKING:
    import lsprotocol.types as types
else:
    import debputy.lsprotocol.types as types


_DISPATCH_RULE = LanguageDispatchRule.new_rule(
    "debputy-config.yaml",
    "debputy-config.yaml",
    [],
    [SecondaryLanguage("yaml", secondary_lookup="basename")],
)


lsp_standard_handler(_DISPATCH_RULE, types.TEXT_DOCUMENT_CODE_ACTION)
lsp_standard_handler(_DISPATCH_RULE, types.TEXT_DOCUMENT_WILL_SAVE_WAIT_UNTIL)


def _initialize_yaml_helper(lint_state: LintState) -> LSPYAMLHelper[None]:
    return LSPYAMLHelper(
        lint_state,
        lint_state.plugin_feature_set.manifest_parser_generator,
        None,
    )


@lint_diagnostics(_DISPATCH_RULE)
async def lint_debputy_config(lint_state: LintState) -> None:
    await generic_yaml_lint(
        lint_state,
        DEBPUTY_CONFIG_PARSER,
        _initialize_yaml_helper,
    )


@lsp_completer(_DISPATCH_RULE)
def deboputy_config_completer(
    ls: "DebputyLanguageServer",
    params: types.CompletionParams,
) -> types.CompletionList | Sequence[types.CompletionItem] | None:
    return generic_yaml_completer(
        ls,
        params,
        DEBPUTY_CONFIG_PARSER,
    )


@lsp_hover(_DISPATCH_RULE)
def debputy_config_hover(
    ls: "DebputyLanguageServer",
    params: types.HoverParams,
) -> types.Hover | None:
    return generic_yaml_hover(ls, params, lambda _: DEBPUTY_CONFIG_PARSER)
